/**
  *  \file planetscentral/dbclient/connection.cc
  *  \brief Database Client: Generic Connection Handling
  */

#include <cerrno>
#include <cstring>
#include <stdexcept>
#include <algorithm>
#include <cctype>
#include <sys/select.h>
#include <unistd.h>
#include "planetscentral/dbclient/connection.h"
#include "server/net.h"
#include "cpluslib/format.h"
#include "int/array.h"
#include "util/misc.h"

/* Timeout used for all transfers */
static const int TIMEOUT = 30;

/*
 *  Debugging: set to "#if 1" to log all traffic
 */
#if 0
#define LOG logTransfer
static void
logTransfer(bool send, const void* data, std::size_t size)
{
    static bool bol = true;
    static int  dir = 2;

    if (!bol && dir != send) {
        write(STDOUT_FILENO, "\n", 1);
        bol = true;
    }
    dir = send;

    const char* p = static_cast<const char*>(data);
    while (size > 0) {
        if (bol) {
            if (send) {
                write(STDOUT_FILENO, "\tSEND: ", 7);
            } else {
                write(STDOUT_FILENO, "\tRECV: ", 7);
            }
        }
        const char* q = static_cast<const char*>(std::memchr(p, '\n', size));
        if (q != 0) {
            // full line
            std::size_t n = q-p+1;
            write(STDOUT_FILENO, p, n);
            size -= n;
            p += n;
            bol = true;
        } else {
            // partial line
            write(STDOUT_FILENO, p, size);
            bol = false;
            size = 0;
        }
    }
}
#else
#define LOG(a,b,c)
#endif

/******************************* Connection ******************************/

/** Create idle connection. */
Connection::Connection()
    : infd(-1),
      outfd(-1),
      buf_fill(0),
      buf_ptr(0),
      tx_fill(0)
{ }

/** Destructor. Close connection if still open. */
Connection::~Connection()
{
    close();
}

/** Connect to server.
    \param host Host to connect to
    \param port Port number or service name to connect to
    \throw runtime_error on error */
void
Connection::connect(string_t host, string_t port)
{
    close();
    int fd = Connect().withAddress(host, port);
    if (fd < 0) {
        throw std::runtime_error(string_t(format("Connection to %s:%s failed: %s",
                                                 host, port, std::strerror(errno))).c_str());
    }
    infd = outfd = fd;
    onConnect();
}

/** Connect to server. Like connect, but retries up to 5 times before failing.
    \param host Host to connect to
    \param port Port number or service name to connect to
    \throw runtime_error on error */
void
Connection::connectRetry(string_t host, string_t port)
{
    int n = 5;
    while (n > 0) {
        --n;
        try {
            connect(host, port);
            break;
        }
        catch (...) {
            if (n == 0) {
                throw;
            }
            sleep(1);
        }
    }
}

/** Reconnect if necessary. Tries to detect connection loss, and
    reconnects to the specified host/port if needed using
    connectRetry().
    \param host Host to connect to
    \param port Port number or service name to connect to
    \throw runtime_error on error
    \retval true Reconnect was necessary
    \retval false Reconnect was not necessary */
bool
Connection::maybeReconnect(string_t host, string_t port)
{
    // Try to detect connection loss.
    if (infd >= 0) {
        // I think I'm connected. Check whether we can read.
        if (buf_ptr >= buf_fill) {
            buf_ptr = buf_fill = 0;
        }
        if (buf_fill >= sizeof(buffer)) {
            // Buffer is full, so we already received more than we wanted.
            // Therefore, we have no place where to put our probed-read byte.
            // This does not happen normally, but it probably means we're still
            // connected.
            return false;
        }

        // Okay, we have room in the buffer. Check whether we can read.
        fd_set readFDs;
        FD_ZERO(&readFDs);
        FD_SET(infd, &readFDs);
        timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 1;
        if (select(infd+1, &readFDs, 0, 0, &timeout) == 1) {
            // Yes, we can read immediately. This means that we either received
            // some data (does not happen normally, but just in case...), or that
            // the connection was lost.
            int n = read(infd, &buffer[buf_ptr], 1);
            if (n == 1) {
                // Read something?
                buf_fill++;
            } else {
                // Nothing read; either an error or connection lost. Close it.
                close();
            }
        } else {
            // Nothing read; connection still open.
        }
    }

    // If connection was lost, reconnect.
    if (infd < 0) {
        connectRetry(host, port);
        return true;
    } else {
        return false;
    }
}

/** Accept connection.
    \param sockfd Socket to accept a connection from
    \retval true Connection accepted
    \retval false accept() returned error */
bool
Connection::accept(int socketfd)
{
    close();
    int fd = ::accept(socketfd, 0, 0);
    if (fd >= 0) {
        infd = outfd = fd;
        return true;
    } else {
        return false;
    }
}

/** Use file descriptor. Both file descriptors can be the same.
    \param infd Input file descriptor
    \param outfd Output file descriptor */
void
Connection::use(int infd, int outfd)
{
    close();
    this->infd = infd;
    this->outfd = outfd;
}

/** Close connection. */
void
Connection::close()
{
    // Close OS handles
    if (infd >= 0 && infd != outfd) {
        ::close(infd);
    }
    if (outfd >= 0) {
        ::close(outfd);
    }
    infd = -1;
    outfd = -1;

    // Reset buffers
    buf_fill = buf_ptr = 0;
    tx_fill = 0;
}

/** Check whether this connection has a file descriptor. Throws if not. */
void
Connection::checkConnected()
{
    if (infd < 0) {
        throw std::runtime_error("Not connected");
    }
}

/** Check whether this connection has a file descriptor. */
bool
Connection::isConnected() const
{
    return infd >= 0;
}

/** Called immediately after obtaining a connection. */
void
Connection::onConnect()
{ }

/** Read object from connection.
    Accepts:
    - "+nnn": string, short form, single line
    - "-nnn": error, generates exception
    - ":nnn": int, short form
    - "$nn\r\n<data>\r\n": string
    - "*nn\r\n<data>": array

    Everything else generates an exception.

    \param acceptShortForm true to accept lines starting with a letter as short form for an array

    \return accepted object (can be null) */
IntValue*
Connection::readObject(bool acceptShortForm)
{
    char c = readChar();
    string_t tmp;
    int32_t n;
    switch (c) {
     case '+':
        // Positive single-line reply. Read a line.
        readLine(tmp);
        return makeStringValue(tmp);
     case '-':
        // Negative single-line reply. Read a line.
        readLine(tmp);
        throw std::runtime_error(tmp.c_str());
     case ':':
        // Integer reply. Read a line.
        n = readInt();
        return makeIntValue(n);
     case '$':
        // Bulk reply.
        n = readInt();
        if (n < 0) {
            // Negative
            return 0;
        } else {
            // Positive
            tmp.clear();
            readBytes(tmp, n);
            while (readChar() != '\n') {
                // skip CRLF after reply
            }
            return makeStringValue(tmp);
        }
     case '*':
        // Multi-bulk reply.
        n = readInt();
        if (n < 0) {
            // Negative
            return 0;
        } else {
            // Positive
            Ptr<IntArrayData> a = new IntArrayData();
            a->addDimension(n);
            for (int32_t i = 0; i < n; ++i) {
                a->content.setNew(i, readObject(false));
            }
            return new IntArray(a);
        }
     default:
        c = std::toupper(static_cast<uint8_t>(c));
        if (acceptShortForm && c >= 'A' && c <= 'Z') {
            // Command line
            Ptr<IntArrayData> a = new IntArrayData();
            tmp.assign(1, c);
            while (1) {
                c = readChar();
                if (c == ' ' || c == '\t' || c == '\r' || c == '\n') {
                    // whitespace
                    if (tmp.size()) {
                        a->content.pushNew(makeStringValue(tmp));
                        tmp.clear();
                    }
                    if (c == '\n') {
                        break;
                    }
                } else {
                    // character
                    if (a->content.size() == 0) {
                        c = std::toupper(static_cast<uint8_t>(c));
                    }
                    tmp.append(1, c);
                }
            }
            a->addDimension(a->content.size());
            return new IntArray(a);
        } else {
            throw std::runtime_error("Protocol syntax error");
        }
    }
}

/** Send object through connection. */
void
Connection::sendObject(IntValue* v)
{
    if (v == 0) {
        send("$-1\r\n", 5);
    } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(v)) {
        // String
        const string_t& s = sv->getValue();
        const string_t head = format("$%d\r\n", s.size());
        send(head.data(), head.size());
        send(s.data(), s.size());
        send("\r\n", 2);
    } else if (IntArray* av = dynamic_cast<IntArray*>(v)) {
        // Array
        Ptr<IntArrayData> a = av->getData();
        sendArray(a->content, a->getDimension(0));
    } else {
        // Anything else. Note that redis does not accept ":1" as an integer
        // within a command, but we accept "$1\r\n1". Therefore, serialize
        // everything in "bulk reply" form.
        const string_t s = IntValue::toStringOf(v, false);
        const string_t head = format("$%d\r\n", s.size());
        send(head.data(), head.size());
        send(s.data(), s.size());
        send("\r\n", 2);
    }
}

/** Send an array.
    \param seg [in] Data segment to send
    \param size [in] Number of elements to send */
void
Connection::sendArray(const IntDataSegment& seg, uint32_t size)
{
    // Send request header
    string_t head = format("*%d\r\n", size);
    send(head.data(), head.size());

    // Send request body
    for (uint32_t i = 0; i < size; ++i) {
        sendObject(seg[i]);
    }
}

/** Send error message. Causes the readObject() at the other end to throw. */
void
Connection::sendError(string_t err)
{
    send("-", 1);
    send(err.data(), err.size());
    send("\r\n", 2);
}

/** Send success message. The readObject() at the other end will read a string. */
void
Connection::sendSuccess(string_t msg)
{
    send("+", 1);
    send(msg.data(), msg.size());
    send("\r\n", 2);
}

/** Flush output. Call after each send so the receiver is guaranteed to
    see the data. */
void
Connection::flush()
{
    if (tx_fill > 0) {
        timedFullWrite(outfd, txbuffer, tx_fill, TIMEOUT);
        LOG(true, txbuffer, tx_fill);
        tx_fill = 0;
    }
}

/** Check for end of input. This may take time to determine, so use
    this only if the next call would have been a possibly-blocking
    readObject() anyway. */
bool
Connection::isAtEnd()
{
    if (buf_ptr >= buf_fill) {
        refillBuffer();
    }
    return (buf_ptr >= buf_fill);
}

/** Refill the buffer. May leave the buffer empty if end-of-input has been reached. */
void
Connection::refillBuffer()
{
    int n = timedRead(infd, buffer, sizeof(buffer), TIMEOUT);
    if (n < 0) {
        throw std::runtime_error("Network timeout/error");
    }
    buf_ptr = 0;
    buf_fill = n;
}

/** Read character. Throws if no character can be read. */
char
Connection::readChar()
{
    if (buf_ptr >= buf_fill) {
        refillBuffer();
        if (buf_fill == 0) {
            throw std::runtime_error("Protocol length error");
        }
        LOG(false, buffer, buf_fill);
    }
    return buffer[buf_ptr++];
}

/** Read integer. Reads an integer and everything up to the next newline.
    Returns the integer. Throws if no "\n" or integer can be read. */
int32_t
Connection::readInt()
{
    int sign = +1;
    int value = 0;
    while (1) {
        char c = readChar();
        if (c == '\n') {
            break;
        } else if (c == '+') {
            sign = +1;
        } else if (c == '-') {
            sign = -1;
        } else if (c >= '0' && c <= '9') {
            value = 10*value + c - '0';
        } else {
            // ignore
        }
    }
    return sign*value;
}

/** Read bytes into a string.
    \param line [out] String to append to
    \param size [in] Number of bytes to read */
void
Connection::readBytes(string_t& line, std::size_t size)
{
    // Reserve space
    line.reserve(line.size() + size);

    // Check bytes available in buffer
    std::size_t avail = buf_fill - buf_ptr;
    if (avail > 0) {
        std::size_t now = std::min(avail, size);
        line.append(&buffer[buf_ptr], now);
        buf_ptr += now;
        size -= now;
    }

    // Read remaining bytes
    while (size > 0) {
        // We can read directly into the buffer, because when size>0 here,
        // buf_ptr==buf_fill, so the buffer is considered empty.
        std::size_t now = std::min(size, sizeof(buffer));
        int n = timedRead(infd, buffer, now, TIMEOUT);
        if (n < 0) {
            throw std::runtime_error("Network timeout/error");
        }
        if (n == 0) {
            throw std::runtime_error("Protocol length error");
        }
        LOG(false, buffer, n);
        line.append(buffer, n);
        size -= n;
    }
}

/** Read a single line. Read up to newline.
    \param line [out] String to append to */
void
Connection::readLine(string_t& line)
{
    while (1) {
        char c = readChar();
        if (c == '\r') {
            // ignore
        } else if (c == '\n') {
            // terminator
            break;
        } else {
            // string character
            line.append(1, c);
        }
    }
}

/** Send raw data.
    \param p [in] Data
    \param n [in] Number of bytes to send */
void
Connection::send(const char* p, std::size_t n)
{
    while (n > 0) {
        // First, try to fill the buffer
        if (tx_fill > 0 || n < sizeof(txbuffer)) {
            std::size_t avail = sizeof(txbuffer) - tx_fill;
            std::size_t now = std::min(n, avail);
            std::memcpy(&txbuffer[tx_fill], p, now);
            tx_fill += now;
            p += now;
            n -= now;
        }

        // If there's data remaining, the buffer must be full, so send it
        if (n != 0) {
            flush();
        }

        // If there's more data remaining than fits in the buffer, send it directly
        if (n >= sizeof(txbuffer)) {
            timedFullWrite(outfd, p, n, TIMEOUT);
            LOG(true, p, n);
            n = 0;
        }
    }
}

/******************************** Globals ********************************/

/** Interpret value as integer result. If the value is actually a
    string, this will attempt to convert it. Otherwise, an error will
    be thrown. */
int32_t
toIntResult(IntValue* v)
{
    if (!v) {
        // Database returned null (null bulk reply)
        return 0;
    } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(v)) {
        // Database returned real integer (redis single-line reply)
        return iv->getValue();
    } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(v)) {
        // Database returned string (redis bulk reply), parse that as integer
        int32_t n;
        if (sv->getValue().empty()) {
            return 0;
        } else if (stringValue(sv->getValue(), n)) {
            return n;
        } else {
            throw std::runtime_error("Protocol type error (expect int, got string)");
        }
    } else {
        throw std::runtime_error("Protocol type error (expect int, got what?)");
    }
}

/** Interpret value as string. */
string_t
toStringResult(IntValue* value)
{
    return IntValue::toStringOf(value, false);
}

/** Interpret value as list of integers.
    \param value [in] Value to check
    \param values [out] Integers will be appended here */
void
toIntListResult(IntValue* value, IntList_t& values)
{
    if (IntArray* a = dynamic_cast<IntArray*>(value)) {
        IntArrayData& d = *a->getData();
        for (uint32_t i = 0, n = d.getDimension(0); i < n; ++i) {
            values.push_back(toIntResult(d.content[i]));
        }
    }
}

/** Interpret value as list of strings.
    \param value [in] Value to check
    \param values [out] Strings will be appended here */
void
toStringListResult(IntValue* value, StringList_t& values)
{
    if (IntArray* a = dynamic_cast<IntArray*>(value)) {
        IntArrayData& d = *a->getData();
        for (uint32_t i = 0, n = d.getDimension(0); i < n; ++i) {
            values.push_back(toStringResult(d.content[i]));
        }
    }
}
