/**
  *  \file planetscentral/talk/cmdpost.cc
  *  \brief Talk: POST commands
  */

#include <stdexcept>
#include <memory>
#include "planetscentral/talk/cmdpost.h"
#include "planetscentral/talk/talkconnection.h"
#include "int/args.h"
#include "cpluslib/string.h"
#include "planetscentral/talk/message.h"
#include "planetscentral/talk/topic.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/forum.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/render.h"
#include "planetscentral/talk/user.h"
#include "planetscentral/talk/notify.h"

PostAccessChecker::PostAccessChecker(TalkConnection& conn)
    : talkConnection(conn),
      lastTopicId(0),
      lastTopicPermitted(false)
{ }

bool
PostAccessChecker::isAllowed(Message& m)
{
    Topic t(m.topic());
    return isAllowed(t) || m.author().get() == talkConnection.getUser();
}

bool
PostAccessChecker::isAllowed(Topic& t)
{
    if (t.getId() != lastTopicId) {
        lastTopicId = t.getId();
        string_t readPermissions = t.readPermissions().get();
        if (readPermissions.empty()) {
            readPermissions = t.forum().readPermissions().get();
        }
        lastTopicPermitted = talkConnection.hasPrivilege(readPermissions);
    }
    return lastTopicPermitted;
}

void
PostAccessChecker::checkMessage(Message& m)
{
    if (!isAllowed(m)) {
        throw std::runtime_error("403 Permission denied");
    }
}

void
PostAccessChecker::checkTopic(Topic& t)
{
    if (!isAllowed(t)) {
        throw std::runtime_error("403 Permission denied");
    }
}

/** POSTNEW fid subj text {USER user|READPERM value|ANSWERPERM value}
    Create new thread and add posting.
    Message attributes:
    - thread = newly allocated
    - time = wall clock time
    - author = authenticated user (or USER attribute for admin)
    - parent, edittime, msgid, rfcheader, subject = null
    - seq = newly allocated
    Thread attributes:
    - subject = as given
    - forum = fid
    - firstpost = new mid
    - readperm, answerperm = as given
    ANSWER: mid */
void
doPostNew(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    enum { HaveUser = 1, HaveReadPerm = 2, HaveAnswerPerm = 4 };
    args.checkArgumentCountAtLeast(3);
    int32_t fid      = toIntResult(args.getNext());
    string_t subject = toStringResult(args.getNext());
    string_t text    = toStringResult(args.getNext());
    string_t user;
    string_t readperm;
    string_t answerperm;
    int32_t have = 0;
    while (args.getNumArgs() > 0) {
        string_t key = strUCase(toStringResult(args.getNext()));
        if (key == "USER") {
            args.checkArgumentCountAtLeast(1);
            user = toStringResult(args.getNext());
            have |= HaveUser;
        } else if (key == "READPERM") {
            args.checkArgumentCountAtLeast(1);
            readperm = toStringResult(args.getNext());
            have |= HaveReadPerm;
        } else if (key == "ANSWERPERM") {
            args.checkArgumentCountAtLeast(1);
            answerperm = toStringResult(args.getNext());
            have |= HaveAnswerPerm;
        } else {
            conn.sendError("400 Syntax error");
            return;
        }
    }

    // Check USER permissions
    if (conn.isAdmin()) {
        if ((have & HaveUser) == 0) {
            conn.sendError("400 Need USER");
            return;
        }
    } else {
        if ((have & HaveUser) != 0 && user != conn.getUser()) {
            conn.sendError("403 USER not allowed");
            return;
        }
        user = conn.getUser();
    }

    // Do it
    ReconnectingGuard g(database_rc);
    Forum f(database_connection, fid);
    if (!f.exists()) {
        conn.sendError("404 No such forum");
        return;
    }
    conn.checkPrivilege(f.writePermissions().get());

    // All preconditions fulfilled, operate!
    const int32_t mid = Message::allocateMessage(database_connection);
    const int32_t tid = Topic::allocateTopic(database_connection);
    const int32_t time = getTime();

    Message msg(database_connection, mid);
    Topic topic(database_connection, tid);
    User u(database_connection, user);

    // Configure message
    msg.topicId().set(tid);
    msg.postTime().set(time);
    msg.author().set(user);
    msg.sequenceNumber().set(++f.lastMessageSequenceNumber());
    msg.text().set(text);
    msg.subject().set(subject);

    // Configure topic
    topic.subject().set(subject);
    topic.forumId().set(fid);
    topic.firstPostingId().set(mid);
    if ((have & HaveReadPerm) != 0) {
        topic.readPermissions().set(readperm);
    }
    if ((have & HaveAnswerPerm) != 0) {
        topic.answerPermissions().set(answerperm);
    }
    topic.lastPost().set(mid);
    topic.lastTime().set(time);

    // Update forum
    f.lastPost().set(mid);
    f.lastTime().set(time);

    // Add message to sets. We can add to f.topics() because topics are all born unsticky.
    topic.messages().add(mid);
    f.messages().add(mid);
    f.topics().add(tid);
    u.postedMessages().add(mid);

    // Notify
    notifyMessage(msg, topic, f);

    // Auto-watch
    if (u.isAutoWatch()) {
        u.watchedTopics().add(tid);
        topic.watchers().add(user);
    }

    IntIntIntValue iv(mid);
    conn.sendObject(&iv);
}

/** POSTREPLY mid subj text {USER user}
    Message attributes:
    - thread = taken from mid
    - parent = mid
    - time = wall clock time
    - author = authenticated user (or USER attribute for admin)
    - edittime, msgid, rfcheader = null
    - subject = as given, null if empty
    - seq = newly allocated
    ANSWER: mid */
void
doPostReply(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    enum { HaveUser = 1 };
    args.checkArgumentCountAtLeast(3);
    int32_t parentId = toIntResult(args.getNext());
    string_t subject = toStringResult(args.getNext());
    string_t text    = toStringResult(args.getNext());
    string_t user;
    int32_t have = 0;
    while (args.getNumArgs() > 0) {
        string_t key = strUCase(toStringResult(args.getNext()));
        if (key == "USER") {
            args.checkArgumentCountAtLeast(1);
            user = toStringResult(args.getNext());
            have |= HaveUser;
        } else {
            conn.sendError("400 Syntax error");
            return;
        }
    }

    // Check USER permissions
    if (conn.isAdmin()) {
        if ((have & HaveUser) == 0) {
            conn.sendError("400 Need USER");
            return;
        }
    } else {
        if ((have & HaveUser) != 0 && user != conn.getUser()) {
            conn.sendError("403 USER not allowed");
            return;
        }
        user = conn.getUser();
    }

    // Do it
    ReconnectingGuard g(database_rc);
    Message parent(database_connection, parentId);
    if (!parent.exists()) {
        conn.sendError("404 Parent message does not exist");
        return;
    }

    const int32_t tid = parent.topicId().get();
    Topic topic(database_connection, tid);

    const int32_t fid = topic.forumId().get();
    Forum f(database_connection, fid);

    string_t answerperm = topic.answerPermissions().get();
    if (answerperm.empty()) {
        answerperm = f.answerPermissions().get();
    }
    if (answerperm.empty()) {
        answerperm = f.writePermissions().get();
    }
    conn.checkPrivilege(answerperm);

    // All preconditions fulfilled, operate!
    const int32_t mid = Message::allocateMessage(database_connection);
    const int32_t time = getTime();
    Message msg(database_connection, mid);
    User u(database_connection, user);

    // Configure message
    msg.topicId().set(tid);
    msg.parentMessageId().set(parentId);
    msg.postTime().set(getTime());
    msg.author().set(user);
    if (subject.empty()) {
        subject = parent.subject().get();
    }
    msg.subject().set(subject);
    msg.sequenceNumber().set(++f.lastMessageSequenceNumber());
    msg.text().set(text);

    // Update topic
    topic.lastPost().set(mid);
    topic.lastTime().set(time);

    // Update forum
    f.lastPost().set(mid);
    f.lastTime().set(time);

    // Add message to sets
    topic.messages().add(mid);
    f.messages().add(mid);
    u.postedMessages().add(mid);

    // Notify
    notifyMessage(msg, topic, f);

    // Auto-watch
    if (u.isAutoWatch()) {
        u.watchedTopics().add(tid);
        topic.watchers().add(user);
    }

    IntIntIntValue iv(mid);
    conn.sendObject(&iv);
}

/** POSTEDIT command mid subj text.
    Message attributes:
    - thread, parent, time, author = unchanged
    - subject, text updated
    - edittime = current time
    - msgid, rfcheader deleted (because this is not a RFC message)
    - seq = updated
    - prevseq, prevmsgid = copied from old seq, msgid (to generate Supersedes header) */
void
doPostEdit(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    enum { HaveUser = 1 };
    args.checkArgumentCount(3);
    int32_t messageId = toIntResult(args.getNext());
    string_t subject  = toStringResult(args.getNext());
    string_t text     = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Message msg(database_connection, messageId);
    if (!msg.exists()) {
        conn.sendError("404 Message not found");
        return;
    }
    if (!conn.isAdmin() && msg.author().get() != conn.getUser()) {
        conn.sendError("403 Not author");
        return;
    }
    if (msg.subject().get() == subject && msg.text().get() == text) {
        conn.sendSuccess("OK, unchanged");
        return;
    }

    // Update message
    const int32_t time = getTime();
    msg.subject().set(subject);
    msg.text().set(text);
    msg.editTime().set(time);

    // Update topic
    Topic topic(database_connection, msg.topicId().get());
    topic.lastTime().set(time);
    if (messageId == topic.firstPostingId().get()) {
        topic.subject().set(subject);
    }

    // Update forum
    Forum f(database_connection, topic.forumId().get());
    f.lastTime().set(time);
    string_t rfcMsgId = msg.rfcMessageId().get();
    msg.previousSequenceNumber().set(msg.sequenceNumber().get());
    msg.previousRfcMessageId().set(rfcMsgId);
    msg.removeRfcMessageId(database_connection, rfcMsgId);
    msg.sequenceNumber().set(++f.lastMessageSequenceNumber());
    msg.rfcMessageId().remove();
    msg.rfcHeaders().remove();

    conn.sendSuccess("OK");
}

/** POSTRENDER command. */
void
doPostRender(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    int32_t messageId = toIntResult(args.getNext());

    // Get options
    RenderState state(conn.getRenderState());
    while (args.getNumArgs() > 0) {
        string_t optUC = strUCase(toStringResult(args.getNext()));
        if (!state.handleOption(optUC, args)) {
            conn.sendError("400 Invalid option");
            return;
        }
    }

    // Do it
    ReconnectingGuard g(database_rc);
    Message msg(database_connection, messageId);
    if (!msg.exists()) {
        conn.sendError("404 Message not found");
        return;
    }
    PostAccessChecker(conn).checkMessage(msg);
    state.setMessageId(messageId);
    IntStringValue sv(state.render(msg.text().get(), database_connection));
    conn.sendObject(&sv);
}

/** POSTMRENDER command. */
void
doPostMRender(TalkConnection& conn, IntArgBlock& args)
{
    // Do it
    ReconnectingGuard g(database_rc);
    IntDataSegment dseg;
    PostAccessChecker checker(conn);
    RenderState state(conn.getRenderState());
    while (args.getNumArgs() > 0) {
        int32_t messageId = toIntResult(args.getNext());
        Message msg(database_connection, messageId);
        if (!msg.exists() || !checker.isAllowed(msg)) {
            dseg.pushNew(0);
        } else {
            state.setMessageId(messageId);
            dseg.pushNew(makeStringValue(state.render(msg.text().get(), database_connection)));
        }
    }
    conn.sendArray(dseg, dseg.size());
}

/** POSTSTAT command. */
void
doPostStat(TalkConnection& conn, IntArgBlock& args)
{
    // Fetch args
    args.checkArgumentCount(1);
    int32_t messageId = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Message msg(database_connection, messageId);
    if (!msg.exists()) {
        conn.sendError("404 Message not found");
        return;
    }
    PostAccessChecker(conn).checkMessage(msg);
    std::auto_ptr<IntValue> result(msg.describe());
    conn.sendObject(result.get());
}

/** POSTMSTAT command. */
void
doPostMStat(TalkConnection& conn, IntArgBlock& args)
{
    // Do it
    ReconnectingGuard g(database_rc);
    IntDataSegment dseg;
    PostAccessChecker checker(conn);
    while (args.getNumArgs() > 0) {
        int32_t messageId = toIntResult(args.getNext());
        Message msg(database_connection, messageId);
        if (!msg.exists() || !checker.isAllowed(msg)) {
            dseg.pushNew(0);
        } else {
            dseg.pushNew(msg.describe());
        }
    }
    conn.sendArray(dseg, dseg.size());
}

/** POSTGET command. */
void
doPostGet(TalkConnection& conn, IntArgBlock& args)
{
    // Fetch args
    args.checkArgumentCount(2);
    int32_t messageId = toIntResult(args.getNext());
    string_t field = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Message msg(database_connection, messageId);
    if (!msg.exists()) {
        conn.sendError("404 Message not found");
        return;
    }
    PostAccessChecker(conn).checkMessage(msg);
    std::auto_ptr<IntValue> result(msg.header().getFieldRaw(field));
    conn.sendObject(result.get());
}

/** POSTRM command. */
void
doPostRemove(TalkConnection& conn, IntArgBlock& args)
{
    // Fetch args
    args.checkArgumentCount(1);
    int32_t messageId = toIntResult(args.getNext());

    // Do it
    int32_t result;
    ReconnectingGuard g(database_rc);
    Message msg(database_connection, messageId);
    if (msg.exists()) {
        // Permission check
        if (!conn.isAdmin() && conn.getUser() != msg.author().get() && !conn.hasPrivilege(msg.topic().forum().deletePermissions().get())) {
            conn.sendError("403 Not author");
            return;
        }

        // Do it
        msg.remove();
        result = 1;
    } else {
        result = 0;
    }
    IntIntIntValue iv(result);
    conn.sendObject(&iv);
}

/** POSTLSNEW command. */
void
doPostListNew(TalkConnection& conn, IntArgBlock& args)
{
    // Fetch args
    args.checkArgumentCount(1);
    int32_t count = toIntResult(args.getNext());

    // Do not check more than this number of postings.
    // This is to avoid checking the whole database for a user who cannot see anything.
    const int32_t MAX_POSTS_TO_CHECK = 200;

    // Do it
    ReconnectingGuard g(database_rc);
    IntDataSegment result;
    int32_t mid = Message::getLastId(database_connection);
    int32_t did = 0;
    PostAccessChecker checker(conn);
    while (mid > 0 && did < MAX_POSTS_TO_CHECK && int32_t(result.size()) < count) {
        // Check this one
        Message m(database_connection, mid);
        if (m.exists() && checker.isAllowed(m)) {
            result.pushNew(makeIntValue(mid));
        }

        // Advance
        --mid;
        ++did;
    }

    // Generate result
    conn.sendArray(result, result.size());
}
